import logging
import os
import sys
from typing import Optional

from torch.utils.tensorboard import SummaryWriter

logger = logging.getLogger(__name__)


class MetricsManager:
    """
    Metrics manager logs for stdout, json and tensorboard

    :param log_dir: (str) Log directory.
    """

    def __init__(self, log_dir: str = "."):
        self.log_dir = log_dir

        self.tensorboard_writer = SummaryWriter(
            os.path.join(self.log_dir, "tensorboard")
        )

    def record_scalar(
        self,
        tag: str,
        scalar: float,
        total_steps: Optional[int] = None,
        tensorboard: bool = False,
    ) -> None:
        print("{}: {:<8.3g}".format(tag, scalar))

        if tensorboard:
            if total_steps is None:
                logger.warning("total_steps argument is required for tensorboard")
            self.tensorboard_writer.add_scalar(tag, scalar, total_steps)

    def dump(self) -> None:
        sys.stdout.flush()
        self.tensorboard_writer.flush()

    def close(self) -> None:
        self.tensorboard_writer.close()



import copy
import logging
import os
import time
from typing import List

import gym
import numpy as np
import torch
from torch import Tensor
from torch.distributions import Distribution
from torch.nn import functional as F

from rl_replicas.experience import Experience
from rl_replicas.metrics_manager import MetricsManager
from rl_replicas.policies import Policy
from rl_replicas.samplers import Sampler
from rl_replicas.utils import (
    bootstrap_rewards_with_last_values,
    compute_values_numpy_list,
    discounted_cumulative_sums,
    gae,
    normalize_tensor,
)
from rl_replicas.value_function import ValueFunction

logger = logging.getLogger(__name__)


class PPO:
    """
    Proximal Policy Optimization (by clipping) with early stopping based on approximate KL divergence

    :param policy: (Policy) Policy.
    :param value_function: (ValueFunction) Value function.
    :param env: (gym.Env) Environment.
    :param sampler: (Sampler) Sampler.
    :param gamma: (float) The discount factor for the cumulative return.
    :param gae_lambda: (float) The factor for trade-off of bias vs variance for GAE.
    :param clip_range: (float) The limit on the likelihood ratio between policies for clipping in the policy objective.
    :param max_kl_divergence: (float) The limit on the KL divergence between policies for early stopping.
    :param num_policy_gradients (int): The number of gradient descent steps to take on policy per epoch.
    :param num_value_gradients (int): The number of gradient descent steps to take on value function per epoch.
    """

    def __init__(
        self,
        policy: Policy,
        value_function: ValueFunction,
        env: gym.Env,
        sampler: Sampler,
        gamma: float = 0.99,
        gae_lambda: float = 0.97,
        clip_range: float = 0.2,
        max_kl_divergence: float = 0.01,
        num_policy_gradients: int = 80,
        num_value_gradients: int = 80,
    ) -> None:
        self.policy = policy
        self.value_function = value_function
        self.env = env
        self.sampler = sampler
        self.gamma = gamma
        self.gae_lambda = gae_lambda
        self.clip_range = clip_range
        self.max_kl_divergence = max_kl_divergence
        self.num_policy_gradients = num_policy_gradients
        self.num_value_gradients = num_value_gradients

        self.old_policy: Policy = copy.deepcopy(self.policy)

    def learn(
        self,
        num_epochs: int = 50,
        batch_size: int = 4000,
        model_saving_interval: int = 4000,
        output_dir: str = ".",
    ) -> None:
        """
        Learn the model

        :param num_epochs: (int) The number of epochs to run and train.
        :param batch_size: (int) The number of steps to run per epoch.
        :param model_saving_interval: (int) The interval steps between model saving.
        :param output_dir: (str) The output directory.
        """
        start_time: float = time.time()
        self.current_total_steps: int = 0
        self.current_total_episodes: int = 0

        os.makedirs(output_dir, exist_ok=True)

        self.metrics_manager: MetricsManager = MetricsManager(output_dir)

        for current_epoch in range(1, num_epochs + 1):
            experience: Experience = self.sampler.sample(batch_size, self.policy)

            episode_returns: List[float] = experience.episode_returns
            episode_lengths: List[int] = experience.episode_lengths

            self.current_total_steps += sum(experience.episode_lengths)
            self.current_total_episodes += sum(experience.episode_dones)

            self.metrics_manager.record_scalar("epoch", current_epoch)
            self.metrics_manager.record_scalar("total_steps", self.current_total_steps)
            self.metrics_manager.record_scalar(
                "total_episodes", self.current_total_episodes
            )

            self.metrics_manager.record_scalar(
                "sampling/average_episode_return",
                float(np.mean(episode_returns)),
                self.current_total_steps,
                tensorboard=True,
            )
            self.metrics_manager.record_scalar(
                "sampling/episode_return_std", float(np.std(episode_returns))
            )
            self.metrics_manager.record_scalar(
                "sampling/max_episode_return", float(np.max(episode_returns))
            )
            self.metrics_manager.record_scalar(
                "sampling/min_episode_return", float(np.min(episode_returns))
            )
            self.metrics_manager.record_scalar(
                "sampling/average_episode_length",
                float(np.mean(episode_lengths)),
                self.current_total_steps,
                tensorboard=True,
            )

            self.train(experience)

            if self.current_total_steps % model_saving_interval == 0:
                model_path: str = os.path.join(output_dir, "model.pt")

                logger.debug("Save model")
                self.save_model(current_epoch, model_path)

            self.metrics_manager.record_scalar("time", time.time() - start_time)

            # Dump all metrics stored in this epoch
            self.metrics_manager.dump()

        self.metrics_manager.close()

    def train(self, experience: Experience) -> None:
        values_numpy_list: np.ndarray = compute_values_numpy_list(
            experience.observations_with_last_observation, self.value_function
        )

        last_values: List[float] = [
            float(episode_values[-1]) for episode_values in values_numpy_list
        ]

        bootstrapped_rewards: List[List[float]] = bootstrap_rewards_with_last_values(
            experience.rewards, experience.episode_dones, last_values
        )

        discounted_returns: List[np.ndarray] = [
            discounted_cumulative_sums(episode_rewards, self.gamma)[:-1]
            for episode_rewards in bootstrapped_rewards
        ]
        flattened_discounted_returns: Tensor = torch.from_numpy(
            np.concatenate(discounted_returns)
        ).float()

        flattened_observations: Tensor = torch.from_numpy(
            np.concatenate(experience.observations)
        ).float()
        flattened_actions: Tensor = torch.from_numpy(
            np.concatenate(experience.actions)
        ).float()

        gaes: List[np.ndarray] = [
            gae(episode_rewards, self.gamma, episode_values, self.gae_lambda)
            for episode_rewards, episode_values in zip(
                bootstrapped_rewards, values_numpy_list
            )
        ]
        flattened_advantages: Tensor = torch.from_numpy(np.concatenate(gaes)).float()
        flattened_advantages = normalize_tensor(flattened_advantages)

        # For logging
        with torch.no_grad():
            policy_dist: Distribution = self.policy(flattened_observations)
            policy_loss_before: Tensor = self.compute_policy_loss(
                flattened_observations, flattened_actions, flattened_advantages
            )
        log_probs_before: Tensor = policy_dist.log_prob(flattened_actions)
        entropies_before: Tensor = policy_dist.entropy()

        # Train policy
        for i in range(self.num_policy_gradients):
            self.train_policy(
                flattened_observations, flattened_actions, flattened_advantages
            )

            approximate_kl_divergence: Tensor = self.compute_approximate_kl_divergence(
                flattened_observations, flattened_actions
            ).detach()
            if approximate_kl_divergence > 1.5 * self.max_kl_divergence:
                logger.info(
                    "Early stopping at update {} due to reaching max KL divergence.".format(
                        i
                    )
                )
                break

        self.old_policy.load_state_dict(self.policy.state_dict())

        # Train value function
        value_function_losses: List[float] = []
        for _ in range(self.num_value_gradients):
            value_function_loss: Tensor = self.train_value_function(
                flattened_observations, flattened_discounted_returns
            )

            value_function_losses.append(value_function_loss.item())

        self.metrics_manager.record_scalar(
            "policy/loss",
            policy_loss_before.item(),
            self.current_total_steps,
            tensorboard=True,
        )
        self.metrics_manager.record_scalar(
            "policy/avarage_entropy",
            torch.mean(entropies_before).item(),
            self.current_total_steps,
            tensorboard=True,
        )
        self.metrics_manager.record_scalar(
            "policy/log_prob_std",
            torch.std(log_probs_before).item(),
            self.current_total_steps,
            tensorboard=True,
        )
        self.metrics_manager.record_scalar(
            "policy/kl_divergence",
            approximate_kl_divergence.item(),
            self.current_total_steps,
            tensorboard=True,
        )
        self.metrics_manager.record_scalar(
            "value_function/average_loss",
            float(np.mean(value_function_losses)),
            self.current_total_steps,
            tensorboard=True,
        )

    def train_policy(
        self,
        flattened_observations: Tensor,
        flattened_actions: Tensor,
        flattened_advantages: Tensor,
    ) -> None:
        policy_loss: Tensor = self.compute_policy_loss(
            flattened_observations, flattened_actions, flattened_advantages
        )

        self.policy.optimizer.zero_grad()
        policy_loss.backward()
        self.policy.optimizer.step()

    def compute_policy_loss(
        self, observations: Tensor, actions: Tensor, advantages: Tensor
    ) -> Tensor:
        policy_dist: Distribution = self.policy(observations)
        log_probs: Tensor = policy_dist.log_prob(actions)

        with torch.no_grad():
            old_policy_dist: Distribution = self.old_policy(observations)
            old_log_probs: Tensor = old_policy_dist.log_prob(actions)

        # Calculate surrogate
        likelihood_ratio: Tensor = torch.exp(log_probs - old_log_probs)
        surrogate: Tensor = likelihood_ratio * advantages

        # Clipping the constraint
        likelihood_ratio_clip: Tensor = torch.clamp(
            likelihood_ratio, min=1 - self.clip_range, max=1 + self.clip_range
        )

        # Calculate surrotate clip
        surrogate_clip: Tensor = likelihood_ratio_clip * advantages

        policy_loss: Tensor = -torch.min(surrogate, surrogate_clip).mean()

        return policy_loss

    def compute_approximate_kl_divergence(
        self, observations: Tensor, actions: Tensor
    ) -> Tensor:
        with torch.no_grad():
            policy_dist: Distribution = self.policy(observations)
            log_probs: Tensor = policy_dist.log_prob(actions)

            old_policy_dist: Distribution = self.old_policy(observations)
            old_log_probs: Tensor = old_policy_dist.log_prob(actions)

        approximate_kl_divergence: Tensor = old_log_probs - log_probs

        return torch.mean(approximate_kl_divergence)

    def train_value_function(
        self, flattened_observations: Tensor, flattened_discounted_returns: Tensor
    ) -> Tensor:
        value_function_loss: Tensor = self.compute_value_function_loss(
            flattened_observations, flattened_discounted_returns
        )

        self.value_function.optimizer.zero_grad()
        value_function_loss.backward()
        self.value_function.optimizer.step()

        return value_function_loss.detach()

    def compute_value_function_loss(
        self, observations: Tensor, discounted_returns: Tensor
    ) -> Tensor:
        values: Tensor = self.value_function(observations)
        squeezed_values: Tensor = torch.squeeze(values, -1)
        value_loss: Tensor = F.mse_loss(squeezed_values, discounted_returns)

        return value_loss

    def save_model(self, epoch: int, model_path: str) -> None:
        """
        Save model

        :param epoch: (int) The current epoch.
        :param model_path: (int) The path to save the model.
        """
        torch.save(
            {
                "epoch": epoch,
                "total_steps": self.current_total_steps,
                "policy_state_dict": self.policy.network.state_dict(),
                "policy_optimizer_state_dict": self.policy.optimizer.state_dict(),
                "value_function_state_dict": self.value_function.network.state_dict(),
                "value_function_optimizer_state_dict": self.value_function.optimizer.state_dict(),
            },
            model_path,
        )



from typing import List, Type

from torch import Tensor, nn


class MLP(nn.Module):
    """
    The multilayer perceptron (MLP)

    :param sizes: (List[int]) The dimensions of all layer(s).
    :param activation_function: (Type[nn.Module]) The activation function.
    :param output_activation_function: (Type[nn.Module]) The output activation function.
    """

    def __init__(
        self,
        sizes: List[int],
        activation_function: Type[nn.Module] = nn.Tanh,
        output_activation_function: Type[nn.Module] = nn.Identity,
    ) -> None:
        super().__init__()

        layers: List[nn.Module] = []
        for i in range(len(sizes) - 1):
            if i < len(sizes) - 2:
                current_activation_function = activation_function
            else:
                current_activation_function = output_activation_function
            layers += [nn.Linear(sizes[i], sizes[i + 1]), current_activation_function()]

        self.network: nn.Module = nn.Sequential(*layers)

    def forward(self, input: Tensor) -> Tensor:
        """
        Forward pass in the MLP

        :param input: (Tensor) The input tensor.
        :return output: (Tensor) The output tensor.
        """
        output: Tensor = self.network(input)
        return output



from abc import ABC, abstractmethod

from rl_replicas.experience import Experience
from rl_replicas.policies import Policy


class Sampler(ABC):
    """
    Base sampler class
    """

    @abstractmethod
    def sample(self, num_samples: int, policy: Policy) -> Experience:
        """
        Sample experience

        :param num_samples: (int) The number of samples to collect.
        :param policy: (Policy) Policy.
        :return: (Experience) Collected experience.
        """
        raise NotImplementedError



import copy
import logging
import os
import time
from typing import Dict, List

import gym
import numpy as np
import torch
from torch import Tensor
from torch.nn import functional as F

from rl_replicas.evaluator import Evaluator
from rl_replicas.experience import Experience
from rl_replicas.metrics_manager import MetricsManager
from rl_replicas.policies import Policy
from rl_replicas.q_function import QFunction
from rl_replicas.replay_buffer import ReplayBuffer
from rl_replicas.samplers import Sampler
from rl_replicas.utils import add_noise_to_get_action, polyak_average

logger = logging.getLogger(__name__)


class TD3:
    """
    Twin Delayed Deep Deterministic Policy Gradient (TD3)

    :param policy: (Policy) Policy.
    :param exploration_policy: (Policy) Exploration policy.
    :param q_function_1: (QFunction) Q function.
    :param q_function_2: (QFunction) Q function.
    :param env: (gym.Env) Environment.
    :param sampler: (Sampler) Sampler.
    :param replay_buffer: (ReplayBuffer) Replay buffer.
    :param evaluator: (Evaluator) Evaluator.
    :param gamma: (float) The discount factor for the cumulative return.
    :param polyak_rho: (float) The interpolation factor in polyak averaging for target networks.
    :param action_noise_scale: (float) The scale of the noise (std) for the policy to explore better.
    :param target_noise_scale: (float) The scale of the smoothing noise (std) for the target policy to exploit harder.
    :param target_noise_clip: (float) The limit for absolute value of the target policy smoothing noise.
    :param policy_delay: (int) The policy will only be updated once every policy_delay times for each update of
        the Q-networks.
    """

    def __init__(
        self,
        policy: Policy,
        exploration_policy: Policy,
        q_function_1: QFunction,
        q_function_2: QFunction,
        env: gym.Env,
        sampler: Sampler,
        replay_buffer: ReplayBuffer,
        evaluator: Evaluator,
        gamma: float = 0.99,
        polyak_rho: float = 0.995,
        action_noise_scale: float = 0.1,
        target_noise_scale: float = 0.2,
        target_noise_clip: float = 0.5,
        policy_delay: int = 2,
    ) -> None:
        self.policy = policy
        self.exploration_policy = exploration_policy
        self.q_function_1 = q_function_1
        self.q_function_2 = q_function_2
        self.env = env
        self.sampler = sampler
        self.replay_buffer = replay_buffer
        self.evaluator = evaluator
        self.gamma = gamma
        self.polyak_rho = polyak_rho
        self.action_noise_scale = action_noise_scale

        self.noised_policy = add_noise_to_get_action(
            self.policy, self.env.action_space, self.action_noise_scale
        )
        self.evaluation_env = gym.make(env.spec.id)
        self.target_policy = copy.deepcopy(self.policy)

        for param in self.target_policy.network.parameters():
            param.requires_grad = False

        self.target_noise_scale = target_noise_scale
        self.target_noise_clip = target_noise_clip
        self.policy_delay = policy_delay

        self.target_q_function_1 = copy.deepcopy(self.q_function_1)
        self.target_q_function_2 = copy.deepcopy(self.q_function_2)

        for param in self.target_q_function_1.network.parameters():
            param.requires_grad = False
        for param in self.target_q_function_2.network.parameters():
            param.requires_grad = False

    def learn(
        self,
        num_epochs: int = 2000,
        batch_size: int = 50,
        minibatch_size: int = 100,
        num_start_steps: int = 10000,
        num_steps_before_update: int = 1000,
        num_train_steps: int = 50,
        num_evaluation_episodes: int = 5,
        evaluation_interval: int = 4000,
        model_saving_interval: int = 4000,
        output_dir: str = ".",
    ) -> None:
        """
        Learn the model

        :param num_epochs: (int) The number of epochs to run and train.
        :param batch_size: (int) The number of steps to run per epoch.
        ;param minibatch_size: (int) The minibatch size for SGD.
        :param num_start_steps: (int) The number of steps for exploration action selection at the beginning.
        :param num_steps_before_update: (int) The number of steps to perform before policy is updated.
        :param num_train_steps: (int) The number of training steps on each epoch.
        :param num_evaluation_episodes: (int) The number of evaluation episodes.
        :param evaluation_interval: (int) The interval steps between evaluation.
        :param model_saving_interval: (int) The interval steps between model saving.
        :param output_dir: (str) The output directory.
        """
        start_time: float = time.time()
        self.current_total_steps: int = 0
        self.current_total_episodes: int = 0

        os.makedirs(output_dir, exist_ok=True)

        self.metrics_manager: MetricsManager = MetricsManager(output_dir)

        for current_epoch in range(1, num_epochs + 1):
            experience: Experience
            if self.current_total_steps < num_start_steps:
                experience = self.sampler.sample(batch_size, self.exploration_policy)
            else:
                experience = self.sampler.sample(batch_size, self.noised_policy)

            self.replay_buffer.add_experience(experience)

            episode_returns: List[float] = experience.episode_returns
            episode_lengths: List[int] = experience.episode_lengths

            self.current_total_steps += sum(experience.episode_lengths)
            self.current_total_episodes += sum(experience.flattened_dones)

            self.metrics_manager.record_scalar("epoch", current_epoch)
            self.metrics_manager.record_scalar("total_steps", self.current_total_steps)
            self.metrics_manager.record_scalar(
                "total_episodes", self.current_total_episodes
            )

            if len(episode_lengths) > 0:
                self.metrics_manager.record_scalar(
                    "sampling/average_episode_return",
                    float(np.mean(episode_returns)),
                    self.current_total_steps,
                    tensorboard=True,
                )
                self.metrics_manager.record_scalar(
                    "sampling/episode_return_std", float(np.std(episode_returns))
                )
                self.metrics_manager.record_scalar(
                    "sampling/max_episode_return", float(np.max(episode_returns))
                )
                self.metrics_manager.record_scalar(
                    "sampling/min_episode_return", float(np.min(episode_returns))
                )
                self.metrics_manager.record_scalar(
                    "sampling/average_episode_length",
                    float(np.mean(episode_lengths)),
                    self.current_total_steps,
                    tensorboard=True,
                )

            if self.current_total_steps >= num_steps_before_update:
                self.train(self.replay_buffer, num_train_steps, minibatch_size)

            if (
                num_evaluation_episodes > 0
                and self.current_total_steps % evaluation_interval == 0
            ):
                evaluation_episode_returns: List[float]
                evaluation_episode_lengths: List[int]
                (
                    evaluation_episode_returns,
                    evaluation_episode_lengths,
                ) = self.evaluator.evaluate(
                    self.policy, self.evaluation_env, num_evaluation_episodes
                )

                self.metrics_manager.record_scalar(
                    "evaluation/average_episode_return",
                    float(np.mean(evaluation_episode_returns)),
                    self.current_total_steps,
                    tensorboard=True,
                )
                self.metrics_manager.record_scalar(
                    "evaluation/episode_return_std",
                    float(np.std(evaluation_episode_returns)),
                )
                self.metrics_manager.record_scalar(
                    "evaluation/max_episode_return",
                    float(np.max(evaluation_episode_returns)),
                )
                self.metrics_manager.record_scalar(
                    "evaluation/min_episode_return",
                    float(np.min(evaluation_episode_returns)),
                )
                self.metrics_manager.record_scalar(
                    "evaluation/average_episode_length",
                    float(np.mean(evaluation_episode_lengths)),
                    self.current_total_steps,
                    tensorboard=True,
                )

            if self.current_total_steps % model_saving_interval == 0:
                model_path: str = os.path.join(output_dir, "model.pt")

                logger.debug("Save model")
                self.save_model(current_epoch, model_path)

            self.metrics_manager.record_scalar("time", time.time() - start_time)

            # Dump all metrics stored in this epoch
            self.metrics_manager.dump()

        self.metrics_manager.close()

    def train(
        self, replay_buffer: ReplayBuffer, num_train_steps: int, minibatch_size: int
    ) -> None:
        policy_losses: List[float] = []
        q_function_1_losses: List[float] = []
        q_function_2_losses: List[float] = []
        all_q_values_1: List[float] = []
        all_q_values_2: List[float] = []

        for train_step in range(num_train_steps):
            minibatch: Dict[str, np.ndarray] = replay_buffer.sample_minibatch(
                minibatch_size
            )

            observations: Tensor = torch.from_numpy(minibatch["observations"]).float()
            actions: Tensor = torch.from_numpy(minibatch["actions"]).float()
            rewards: Tensor = torch.from_numpy(minibatch["rewards"]).float()
            next_observations: Tensor = torch.from_numpy(
                minibatch["next_observations"]
            ).float()
            dones: Tensor = torch.from_numpy(minibatch["dones"]).int()

            # For logging
            with torch.no_grad():
                q_values_1: Tensor = self.q_function_1(observations, actions)
                q_values_2: Tensor = self.q_function_2(observations, actions)
            all_q_values_1.extend(q_values_1.tolist())
            all_q_values_2.extend(q_values_2.tolist())

            targets: Tensor = self.compute_targets(next_observations, rewards, dones)

            q_function_1_loss: Tensor = self.train_q_function(
                self.q_function_1, observations, actions, targets
            )
            q_function_2_loss: Tensor = self.train_q_function(
                self.q_function_2, observations, actions, targets
            )
            q_function_1_losses.append(q_function_1_loss.item())
            q_function_2_losses.append(q_function_2_loss.item())

            if train_step % self.policy_delay == 0:
                policy_loss: Tensor = self.train_policy(observations)
                policy_losses.append(policy_loss.item())

                # Update targets
                polyak_average(
                    self.policy.network.parameters(),
                    self.target_policy.network.parameters(),
                    self.polyak_rho,
                )
                polyak_average(
                    self.q_function_1.network.parameters(),
                    self.target_q_function_1.network.parameters(),
                    self.polyak_rho,
                )
                polyak_average(
                    self.q_function_2.network.parameters(),
                    self.target_q_function_2.network.parameters(),
                    self.polyak_rho,
                )

        self.metrics_manager.record_scalar(
            "policy/average_loss",
            float(np.mean(policy_losses)),
            self.current_total_steps,
            tensorboard=True,
        )
        self.metrics_manager.record_scalar(
            "q-function_1/average_loss",
            float(np.mean(q_function_1_losses)),
            self.current_total_steps,
            tensorboard=True,
        )
        self.metrics_manager.record_scalar(
            "q-function_2/average_loss",
            float(np.mean(q_function_2_losses)),
            self.current_total_steps,
            tensorboard=True,
        )

        self.metrics_manager.record_scalar(
            "q-function_1/avarage_q-value",
            float(np.mean(all_q_values_1)),
            self.current_total_steps,
            tensorboard=True,
        )
        self.metrics_manager.record_scalar(
            "q-function_1/max_q-value", float(np.max(all_q_values_1))
        )
        self.metrics_manager.record_scalar(
            "q-function_1/min_q-value", float(np.min(all_q_values_1))
        )
        self.metrics_manager.record_scalar(
            "q-function_2/avarage_q-value",
            float(np.mean(all_q_values_2)),
            self.current_total_steps,
            tensorboard=True,
        )
        self.metrics_manager.record_scalar(
            "q-function_2/max_q-value", float(np.max(all_q_values_2))
        )
        self.metrics_manager.record_scalar(
            "q-function_2/min_q-value", float(np.min(all_q_values_2))
        )

    def train_policy(self, observations: Tensor) -> Tensor:
        # Freeze Q-networks
        for param in self.q_function_1.network.parameters():
            param.requires_grad = False
        for param in self.q_function_2.network.parameters():
            param.requires_grad = False

        policy_actions: Tensor = self.policy(observations)
        policy_q_values: Tensor = self.q_function_1(observations, policy_actions)

        policy_loss: Tensor = -torch.mean(policy_q_values)

        self.policy.optimizer.zero_grad()
        policy_loss.backward()
        self.policy.optimizer.step()

        # Unfreeze Q-networks
        for param in self.q_function_1.network.parameters():
            param.requires_grad = True
        for param in self.q_function_2.network.parameters():
            param.requires_grad = True

        return policy_loss.detach()

    def compute_targets(
        self, next_observations: Tensor, rewards: Tensor, dones: Tensor
    ) -> Tensor:
        with torch.no_grad():
            next_actions: Tensor = self.target_policy(next_observations)
        epsilon: Tensor = self.target_noise_scale * torch.randn_like(next_actions)
        epsilon = torch.clamp(epsilon, -self.target_noise_clip, self.target_noise_clip)
        next_actions = next_actions + epsilon
        action_limit: float = self.env.action_space.high[0]
        next_actions = torch.clamp(next_actions, -action_limit, action_limit)

        with torch.no_grad():
            target_q_values_1: Tensor = self.target_q_function_1(
                next_observations, next_actions
            )
            target_q_values_2: Tensor = self.target_q_function_2(
                next_observations, next_actions
            )
        target_q_values: Tensor = torch.min(target_q_values_1, target_q_values_2)

        targets: Tensor = rewards + self.gamma * (1 - dones) * target_q_values

        return targets

    def train_q_function(
        self,
        q_function: QFunction,
        observations: Tensor,
        actions: Tensor,
        targets: Tensor,
    ) -> Tensor:
        q_values: Tensor = q_function(observations, actions)

        q_function_loss: Tensor = F.mse_loss(q_values, targets)

        q_function.optimizer.zero_grad()
        q_function_loss.backward()
        q_function.optimizer.step()

        return q_function_loss.detach()

    def save_model(self, current_epoch: int, model_path: str) -> None:
        """
        Save model

        :param current_epoch: (int) The current epoch.
        :param model_path: (int) The path to save the model.
        """
        torch.save(
            {
                "epoch": current_epoch,
                "total_steps": self.current_total_steps,
                "policy_state_dict": self.policy.network.state_dict(),
                "policy_optimizer_state_dict": self.policy.optimizer.state_dict(),
                "target_policy_state_dict": self.target_policy.network.state_dict(),
                "q_function_1_state_dict": self.q_function_1.network.state_dict(),
                "q_function_1_optimizer_state_dict": self.q_function_1.optimizer.state_dict(),
                "target_q_function_1_state_dict": self.target_q_function_1.network.state_dict(),
                "q_function_2_state_dict": self.q_function_2.network.state_dict(),
                "q_function_2_optimizer_state_dict": self.q_function_2.optimizer.state_dict(),
                "target_q_function_2_state_dict": self.target_q_function_2.network.state_dict(),
            },
            model_path,
        )



from abc import abstractmethod

import numpy as np
import torch
from torch import Tensor
from torch.distributions import Distribution

from rl_replicas.policies.policy import Policy


class StochasticPolicy(Policy):
    """
    Abstract base class for stochastic policies
    """

    @abstractmethod
    def forward(self, observation: Tensor) -> Distribution:
        """
        Forward pass in policy

        :param observation: (Tensor) Observation from the environment.
        :return: (Distribution) The distribution of action(s).
        """
        raise NotImplementedError

    def get_action_tensor(self, observation: Tensor) -> Tensor:
        with torch.no_grad():
            distribution: Distribution = self.forward(observation)

        action: Tensor = distribution.sample()

        return action

    def get_action_numpy(self, observation: np.ndarray) -> np.ndarray:
        observation_tensor: Tensor = torch.from_numpy(observation).float()
        with torch.no_grad():
            distribution: Distribution = self.forward(observation_tensor)

        action: Tensor = distribution.sample()

        return action.detach().numpy()



from operator import itemgetter
from typing import Dict, List

import numpy as np

from rl_replicas.experience import Experience


class ReplayBuffer:
    """
    Replay buffer for off-policy algorithms

    :param buffer_size: (int) The size of the replay buffer.
    """

    def __init__(self, buffer_size: int = int(1e6)) -> None:
        self.buffer_size = buffer_size

        self.current_size: int = 0
        self.observations: List[np.ndarray] = []
        self.actions: List[np.ndarray] = []
        self.rewards: List[float] = []
        self.next_observations: List[np.ndarray] = []
        self.dones: List[bool] = []

    def add_experience(self, experience: Experience) -> None:
        """
        Add experience

        :param experience: (Experience) Experience.
        """
        self.observations.extend(experience.flattened_observations)
        self.actions.extend(experience.flattened_actions)
        self.rewards.extend(experience.flattened_rewards)
        self.next_observations.extend(experience.flattened_next_observations)
        self.dones.extend(experience.flattened_dones)

        self.current_size += len(experience.flattened_observations)

        if self.current_size > self.buffer_size:
            num_exceeded_experinece: int = self.current_size - self.buffer_size

            del self.observations[:num_exceeded_experinece]
            del self.actions[:num_exceeded_experinece]
            del self.rewards[:num_exceeded_experinece]
            del self.next_observations[:num_exceeded_experinece]
            del self.dones[:num_exceeded_experinece]

            self.current_size -= num_exceeded_experinece

    def sample_minibatch(self, minibatch_size: int = 32) -> Dict[str, np.ndarray]:
        """
        Sample minibatch

        :param minibatch_size: (int) The number of transitions to be sampled.
        :return: (Dict[str, np.ndarray]) Sampled transitions.
        """
        indices = np.random.randint(0, self.current_size, minibatch_size)

        sampled_observations: np.ndarray = np.vstack(
            itemgetter(*indices)(self.observations)
        )
        sampled_actions: np.ndarray = np.vstack(itemgetter(*indices)(self.actions))
        sampled_rewards: np.ndarray = np.asarray(itemgetter(*indices)(self.rewards))
        sampled_next_observations: np.ndarray = np.vstack(
            itemgetter(*indices)(self.next_observations)
        )
        sampled_dones: np.ndarray = np.asarray(itemgetter(*indices)(self.dones))

        minibatch: Dict[str, np.ndarray] = {
            "observations": sampled_observations,
            "actions": sampled_actions,
            "rewards": sampled_rewards,
            "next_observations": sampled_next_observations,
            "dones": sampled_dones,
        }

        return minibatch



from torch import Tensor, nn
from torch.optim import Optimizer


class ValueFunction(nn.Module):
    """
    The Value Function

    :param optimizer: (Optimizer) The optimizer.
    :param network: (nn.Module) The network.
    """

    def __init__(self, network: nn.Module, optimizer: Optimizer) -> None:
        super().__init__()

        self.network = network
        self.optimizer = optimizer

    def forward(self, observation: Tensor) -> Tensor:
        """
        Forward pass in the value function

        :param observation: (Tensor) Observation(s) from the environment.
        :return value: (Tensor) The value(s).
        """
        value: Tensor = self.network(observation)

        return value



"""
Reinfrocement Learning Algorithms
"""

from rl_replicas.algorithms.ddpg import DDPG
from rl_replicas.algorithms.ppo import PPO
from rl_replicas.algorithms.td3 import TD3
from rl_replicas.algorithms.trpo import TRPO
from rl_replicas.algorithms.vpg import VPG



from typing import List, Optional

import numpy as np


class Experience:
    """
    Experience

    N: The number of episodes.
    L: The length of each episode (it may vary).
    A^*: The shape of single action step.
    O^*: The shape of single observation step.

    :param observations: (Optional[List[List[np.ndarray]]]) A nested list of shape (N, L, O^*).
    :param actions: (Optional[List[List[np.ndarray]]]) A nested list of shape (N, L, A^*).
    :param rewards: (Optional[List[List[float]]]) A nested list of shape (N, L).
    :param last_observations: (Optional[List[np.ndarray]]) A list of np.ndarray (N, O^*).
    :param dones: (Optional[List[bool]]) A nested list shape (N, L).
    :param episode_returns: (Optional[List[float]]) A list with length (N).
    :param episode_lengths: (Optional[List[int]]) A list with length (N).
    """

    def __init__(
        self,
        observations: Optional[List[List[np.ndarray]]] = None,
        actions: Optional[List[List[np.ndarray]]] = None,
        rewards: Optional[List[List[float]]] = None,
        last_observations: Optional[List[np.ndarray]] = None,
        dones: Optional[List[List[bool]]] = None,
        episode_returns: Optional[List[float]] = None,
        episode_lengths: Optional[List[int]] = None,
    ):
        self.observations = observations if observations else []
        self.actions = actions if actions else []
        self.rewards = rewards if rewards else []
        self.last_observations = last_observations if last_observations else []
        self.dones = dones if dones else []
        self.episode_returns = episode_returns if episode_returns else []
        self.episode_lengths = episode_lengths if episode_lengths else []

    @property
    def observations_with_last_observation(self) -> List[List[np.ndarray]]:
        return [
            observations + [last_observation]
            for observations, last_observation in zip(
                self.observations, self.last_observations
            )
        ]

    @property
    def next_observations(self) -> List[List[np.ndarray]]:
        return [
            observations[1:] + [last_observation]
            for observations, last_observation in zip(
                self.observations, self.last_observations
            )
        ]

    @property
    def episode_dones(self) -> List[bool]:
        return [dones_per_episode[-1] for dones_per_episode in self.dones]

    @property
    def flattened_observations(self) -> List[np.ndarray]:
        return [
            observation
            for observations_per_episode in self.observations
            for observation in observations_per_episode
        ]

    @property
    def flattened_actions(self) -> List[np.ndarray]:
        return [
            action
            for actions_per_episode in self.actions
            for action in actions_per_episode
        ]

    @property
    def flattened_rewards(self) -> List[float]:
        return [
            reward
            for rewards_per_episode in self.rewards
            for reward in rewards_per_episode
        ]

    @property
    def flattened_next_observations(self) -> List[np.ndarray]:
        return [
            next_observation
            for next_observations_per_episode in self.next_observations
            for next_observation in next_observations_per_episode
        ]

    @property
    def flattened_dones(self) -> List[bool]:
        return [done for dones_per_episode in self.dones for done in dones_per_episode]



import torch
from torch import Tensor, nn
from torch.optim import Optimizer


class QFunction(nn.Module):
    """
    Q Function

    :param network: (nn.Module) Network.
    :param optimizer: (Optimizer) Optimizer.
    """

    def __init__(self, network: nn.Module, optimizer: Optimizer) -> None:
        super().__init__()

        self.network = network
        self.optimizer = optimizer

    def forward(self, observation: Tensor, action: Tensor) -> Tensor:
        """
        Forward pass in the Q function

        :param observation: (Tensor) The observation(s) from the environment.
        :param action: (Tensor) Action(s) taken.
        :return squeezeed_q_value: (Tensor) The Q value(s).
        """
        input: Tensor = torch.cat([observation, action], -1)
        q_value: Tensor = self.network(input)
        squeezeed_q_value: Tensor = torch.squeeze(q_value, -1)

        return squeezeed_q_value



import torch
from torch import Tensor, nn
from torch.distributions import Independent, Normal
from torch.optim import Optimizer

from rl_replicas.policies.stochastic_policy import StochasticPolicy


class GaussianPolicy(StochasticPolicy):
    """
    Gaussian policy

    :param network: (nn.Module) Network.
    :param optimizer: (Optimizer) Optimizer.
    :param log_std: (Tensor) The standard deviation of the distribution.
    """

    def __init__(self, network: nn.Module, optimizer: Optimizer, log_std: nn.Parameter):
        super().__init__()

        self.network = network
        self.optimizer = optimizer
        self.log_std = log_std

    def forward(self, observation: Tensor) -> Independent:
        """
        Forward pass in policy

        :param observation: (Tensor) Observation(s) from the environment
        :return: (Independent) The normal (also called Gaussian) distribution of action(s).
        """
        mean: Tensor = self.network(observation)
        std = torch.exp(self.log_std)
        # Use Independent for changing the shape of the result of log_prob()
        distribution: Independent = Independent(Normal(loc=mean, scale=std), 1)

        return distribution



from torch import Tensor, nn
from torch.distributions import Categorical
from torch.optim import Optimizer

from rl_replicas.policies.stochastic_policy import StochasticPolicy


class CategoricalPolicy(StochasticPolicy):
    """
    Categorical policy

    :param network: (nn.Module) Network.
    :param optimizer: (Optimizer) Optimizer.
    """

    def __init__(self, network: nn.Module, optimizer: Optimizer):
        super().__init__()

        self.network = network
        self.optimizer = optimizer

    def forward(self, observation: Tensor) -> Categorical:
        """
        Forward pass in policy

        :param observation: (Tensor) Observation of the environment.
        :return: (Categorical) The distribution of action(s).
        """
        logits: Tensor = self.network(observation)
        distribution: Categorical = Categorical(logits=logits)

        return distribution



import logging
from typing import Callable, Iterable, List, Tuple

import numpy as np
import torch
from torch import Tensor
from torch.optim import Optimizer
from typing_extensions import TypedDict

logger = logging.getLogger(__name__)

State = TypedDict(
    "State",
    {
        "max_constraint": float,
        "n_conjugate_gradients": int,
        "max_backtracks": int,
        "backtrack_ratio": float,
        "hvp_damping_coefficient": float,
    },
)


class ConjugateGradientOptimizer(Optimizer):
    """
    Performs constrained optimization via backtracking line search

    The search direction is computed using a conjugate gradient algorithm,
    which gives x = H^{-1}g, where H is a second order approximation of the
    constraint and g is the gradient of the loss function.

    :param params: (Iterable) A iterable of parameters to optimize.
    :param max_constraint: (float) The maximum constraint value.
    :param n_conjugate_gradients: (int) The number of conjugate gradient iterations used to calculate H^-1 g.
    :param max_backtracks: (int) The max number of iterations for backtrack linesearch.
    :param backtrack_ratio: (float) The backtrack ratio for backtracking line search.
    :param hvp_damping_coefficient: (float) The coefficient for numerical stability, should be smallish.
        Adjusts Hessian-vector product calculation: H -> H + hvp_damping_coefficient*I.
    """

    def __init__(
        self,
        params: Iterable[Tensor],
        max_constraint: float = 0.01,
        n_conjugate_gradients: int = 10,
        max_backtracks: int = 15,
        backtrack_ratio: float = 0.8,
        hvp_damping_coefficient: float = 1e-5,
    ):
        # Initializing defaults is required
        defaults: dict = {}
        super().__init__(params, defaults)
        self.max_constraint = max_constraint
        self.n_conjugate_gradients = n_conjugate_gradients
        self.max_backtracks = max_backtracks
        self.backtrack_ratio = backtrack_ratio
        self.hvp_damping_coefficient = hvp_damping_coefficient

    def step(self, loss_function: Callable, kl_divergence_function: Callable) -> None:
        """
        Performs a single optimization step

        :param loss_function: (Callable) A function to compute the loss.
        :param kl_divergence_function: (Callable) A function to compute the kl divergence.
        """
        # Collect trainable parameters and gradients
        params: List[Tensor] = []
        grads: List[Tensor] = []
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is not None:
                    params.append(p)
                    grads.append(p.grad.reshape(-1))
        flat_loss_grads: Tensor = torch.cat(grads)

        # Build Hessian-vector-product function
        hessian_vector_product_function = self._build_hessian_vector_product(
            kl_divergence_function, params
        )

        # Compute step direction
        step_direction = self._conjugate_gradient(
            hessian_vector_product_function, flat_loss_grads
        )

        # Replace nan with 0.0
        step_direction[step_direction.ne(step_direction)] = 0.0

        # Compute step size
        step_size = np.sqrt(
            2.0
            * self.max_constraint
            * (
                1.0
                / (
                    torch.dot(
                        step_direction, hessian_vector_product_function(step_direction)
                    )
                    + 1e-8
                )
            )
        )

        if np.isnan(step_size):
            step_size = 1.0

        descent_step = step_size * step_direction

        # Update parameters using backtracking line search
        self._backtracking_line_search(
            params, descent_step, loss_function, kl_divergence_function
        )

    @property
    def state(self) -> State:
        """
        :return: (State) The hyper-parameters of the optimizer.
        """
        return {
            "max_constraint": self.max_constraint,
            "n_conjugate_gradients": self.n_conjugate_gradients,
            "max_backtracks": self.max_backtracks,
            "backtrack_ratio": self.backtrack_ratio,
            "hvp_damping_coefficient": self.hvp_damping_coefficient,
        }

    @state.setter
    def state(self, state: State) -> None:
        self.max_constraint = state.get("max_constraint", 0.01)
        self.n_conjugate_gradients = state.get("n_conjugate_gradients", 10)
        self.max_backtracks = state.get("max_backtracks", 15)
        self.backtrack_ratio = state.get("backtrack_ratio", 0.8)
        self.hvp_damping_coefficient = state.get("hvp_damping_coefficient", 1e-5)

    def __setstate__(self, state: dict) -> None:
        """
        Restore the optimizer state

        :param state: (dict) State dictionary.
        """
        if "hvp_damping_coefficient" not in state["state"]:
            logger.warning("Resuming ConjugateGradientOptimizer with lost state.")
        # Set the fields manually so that the setter gets called.
        self.state = state["state"]
        self.param_groups = state["param_groups"]

    def _build_hessian_vector_product(
        self, hessian_target_vector_function: Callable, params: List[Tensor]
    ) -> Callable:
        param_shapes: List[torch.Size] = [p.shape or torch.Size([1]) for p in params]
        hessian_target_vector = hessian_target_vector_function()
        hessian_target_vector_grads: Tuple[Tensor, ...] = torch.autograd.grad(
            hessian_target_vector, params, create_graph=True
        )

        def _eval(vector: Tensor) -> Tensor:
            """
            The evaluation function

            :param vector (Tensor): The vector to be multiplied with Hessian.
            :return: (Tensor) The product of Hessian of function f and v.
            """
            unflatten_vector: Tensor = self.unflatten_tensors(vector, param_shapes)

            assert len(hessian_target_vector_grads) == len(unflatten_vector)
            grad_vector_product_list: List[Tensor] = []
            for g, x in zip(hessian_target_vector_grads, unflatten_vector):
                single_grad_vector_product = torch.sum(g * x)
                grad_vector_product_list.append(single_grad_vector_product)

            grad_vector_product = torch.sum(torch.stack(grad_vector_product_list))

            hvp: List[Tensor] = list(
                torch.autograd.grad(grad_vector_product, params, retain_graph=True)
            )
            for i, (hx, p) in enumerate(zip(hvp, params)):
                if hx is None:
                    hvp[i] = torch.zeros_like(p)

            flat_output: Tensor = torch.cat([h.reshape(-1) for h in hvp])
            return flat_output + self.hvp_damping_coefficient * vector

        return _eval

    def _conjugate_gradient(
        self,
        hessian_vector_product_function: Callable,
        b: Tensor,
        residual_tol: float = 1e-10,
    ) -> Tensor:
        """
        Use Conjugate Gradient iteration to solve Ax = b. Demmel p 312

        :param hessian_vector_product_function: (Callable) A function to compute Hessian vector product.
        :param b: (Tensor) The right hand side of the equation to solve.
        :param residual_tol: (float) Tolerence for convergence.

        :return: (Tensor) Solution x* for equation Ax = b.
        """
        x = torch.zeros_like(b)

        r = b.clone()
        p = b.clone()
        rdotr = torch.dot(r, r)

        for _ in range(self.n_conjugate_gradients):
            z = hessian_vector_product_function(p)
            v = rdotr / torch.dot(p, z)
            x += v * p
            r -= v * z
            newrdotr = torch.dot(r, r)
            mu = newrdotr / rdotr
            p = r + mu * p

            rdotr = newrdotr
            if rdotr < residual_tol:
                break
        return x

    def _backtracking_line_search(
        self,
        params: List[Tensor],
        descent_step: float,
        loss_function: Callable,
        constraint_function: Callable,
    ) -> None:
        previous_params: List[Tensor] = [p.clone() for p in params]
        ratio_list: np.ndarray = self.backtrack_ratio ** np.arange(self.max_backtracks)
        loss_before: Tensor = loss_function()

        param_shapes: List[torch.Size] = [p.shape or torch.Size([1]) for p in params]
        descent_step_list: List[Tensor] = self.unflatten_tensors(
            descent_step, param_shapes
        )
        assert len(descent_step_list) == len(params)

        for ratio in ratio_list:
            for step, previous_param, param in zip(
                descent_step_list, previous_params, params
            ):
                step = ratio * step
                new_param = previous_param.data - step
                param.data = new_param.data

            new_loss = loss_function()
            constraint = constraint_function()

            if new_loss < loss_before and constraint <= self.max_constraint:
                break

        if (
            torch.isnan(new_loss)
            or torch.isnan(constraint)
            or new_loss >= loss_before
            or constraint >= self.max_constraint
        ):
            logger.warning("Line search condition violated. Rejecting the step.")
            if torch.isnan(new_loss):
                logger.debug("Violated because loss is NaN")
            if torch.isnan(constraint):
                logger.debug("Violated because constraint is NaN")
            if new_loss >= loss_before:
                logger.debug("Violated because loss not improving")
            if constraint >= self.max_constraint:
                logger.debug("Violated because constraint is violated")

            for previous_param, param in zip(previous_params, params):
                param.data = previous_param.data

    def unflatten_tensors(
        self, flattened: np.ndarray, tensor_shapes: List[torch.Size]
    ) -> List[Tensor]:
        """
        Unflatten a flattened tensors into a list of tensors

        :param flattened: (np.ndarray) Flattened tensors.
        :param tensor_shapes: (List[torch.Size]) Tensor shapes.
        :return: (List[np.ndarray]) Unflattened list of tensors.
        """
        tensor_sizes = list(map(np.prod, tensor_shapes))
        indices = np.cumsum(tensor_sizes)[:-1]

        return [
            np.reshape(pair[0], pair[1])
            for pair in zip(np.split(flattened, indices), tensor_shapes)
        ]



import random
from typing import Iterable, List

import numpy as np
import scipy.signal
import torch
from gym import Space
from torch import Tensor, nn

from rl_replicas.policies.policy import Policy
from rl_replicas.value_function import ValueFunction


def discounted_cumulative_sums(vector: np.ndarray, discount: float) -> np.ndarray:
    """
    Compute discounted cumulative sums of vector

    :param vector: (np.ndarray) A target vector.
    e.g. [x0,
          x1,
          x2]
    :param discount: (float) The discount factor for the cumulative return.
    :return: (np.ndarray) The discounted cumulative sums of a vector.
    e.g. [x0 + discount * x1 + discount^2 * x2,
          x1 + discount * x2,
          x2]
    """
    return scipy.signal.lfilter([1], [1, -discount], vector[::-1], axis=0)[::-1]


def gae(
    rewards: np.ndarray, gamma: float, values: np.ndarray, gae_lambda: float
) -> np.ndarray:
    """
    Compute Generalized Advantage Estimation (GAE)

    :param rewards: (np.ndarray) Rewards for all states.
    :param gamma: (float) The discount factor for the cumulative return.
    :param values: (np.ndarray) Values for all states.
    :param gae_lambda: (float) A smoothing parameter for reducing the variance.
    :return gaes: (np.ndarray) The GAEs for all states.
    """
    deltas: np.ndarray = rewards[:-1] + gamma * values[1:] - values[:-1]
    gaes: np.ndarray = discounted_cumulative_sums(deltas, gamma * gae_lambda)

    return gaes


def polyak_average(
    params: Iterable[nn.Parameter], target_params: Iterable[nn.Parameter], rho: float
) -> None:
    """
    Perform Polyak averaging on target_params using params

    :param params: (Iterable[torch.nn.Parameter]) The parameters to use to update the target params.
    :param target_params: (Iterable[torch.nn.Parameter]) The parameters to update.
    :param rho: (float) The coefficient for polyak averaging (between 0 and 1).
    """
    with torch.no_grad():
        for param, target_param in zip(params, target_params):
            target_param.data.copy_(
                torch.tensor(rho) * target_param.data
                + torch.tensor(1.0 - rho) * param.data
            )


def compute_values_numpy_list(
    observations_with_last_observation: List[List[np.ndarray]],
    value_function: ValueFunction,
) -> np.ndarray:
    values_numpy_list: List[np.ndarray] = []
    with torch.no_grad():
        for (
            episode_observations_with_last_observation
        ) in observations_with_last_observation:
            episode_observations_with_last_observation_tensor = torch.from_numpy(
                np.concatenate([episode_observations_with_last_observation])
            ).float()
            values_numpy_list.append(
                value_function(episode_observations_with_last_observation_tensor)
                .flatten()
                .numpy()
            )
    return values_numpy_list


def bootstrap_rewards_with_last_values(
    rewards: List[List[float]], episode_dones: List[bool], last_values: List[float]
) -> List[List[float]]:
    bootstrapped_rewards: List[List[float]] = []

    for episode_rewards, episode_done, last_value in zip(
        rewards, episode_dones, last_values
    ):
        episode_bootstrapped_rewards: List[float]
        if episode_done:
            episode_bootstrapped_rewards = episode_rewards + [0]
        else:
            episode_bootstrapped_rewards = episode_rewards + [last_value]
        bootstrapped_rewards.append(episode_bootstrapped_rewards)

    return bootstrapped_rewards


def normalize_tensor(vector: Tensor) -> Tensor:
    normalized_vector = (vector - torch.mean(vector)) / torch.std(vector)
    return normalized_vector


def add_noise_to_get_action(
    policy: Policy, action_space: Space, action_noise_scale: float
) -> Policy:
    noised_policy: Policy = _NoisedPolicy(policy, action_space, action_noise_scale)

    return noised_policy


class _NoisedPolicy(Policy):
    def __init__(
        self, base_policy: Policy, action_space: Space, action_noise_scale: float
    ):
        super().__init__()

        self.base_policy = base_policy
        self.action_space = action_space
        self.action_noise_scale = action_noise_scale

        self.action_limit = action_space.high[0]
        self.action_size = action_space.shape[0]

    def get_action_tensor(self, observation: Tensor) -> Tensor:
        action = self.base_policy.get_action_tensor(observation)
        action += self.action_noise_scale * torch.randn(self.action_size)
        action = torch.clip(action, -self.action_limit, self.action_limit)

        return action

    def get_action_numpy(self, observation: np.ndarray) -> np.ndarray:
        action = self.base_policy.get_action_numpy(observation)
        action += self.action_noise_scale * np.random.randn(self.action_size)
        action = np.clip(action, -self.action_limit, self.action_limit)

        return action


def set_seed_for_libraries(seed: int) -> None:
    """
    Set seed for random, numpy and torch.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



import logging
from typing import List, Optional

import gym
import numpy as np

from rl_replicas.experience import Experience
from rl_replicas.policies import Policy
from rl_replicas.samplers import Sampler

logger = logging.getLogger(__name__)


class BatchSampler(Sampler):
    """
    Batch sampler

    :param env: (gym.Env) Environment.
    :param seed: (int) Seed.
    :param is_continuous: (bool) If true, observation is retained and
        it samples experiences continuously across sample function calls.
    """

    def __init__(
        self, env: gym.Env, seed: Optional[int] = None, is_continuous: bool = False
    ):
        self.env = env
        self.seed = seed
        self.is_continuous = is_continuous

        self.observation: Optional[np.ndarray] = None

    def sample(self, num_samples: int, policy: Policy) -> Experience:
        """
        Sample experience

        :param num_samples: (int) The number of samples to collect.
        :param policy: (int) Policy.
        :return: (Experience) Sampled experience.
        """
        experience: Experience = Experience()

        # Variables on each episode
        episode_observations: List[np.ndarray] = []
        episode_actions: List[np.ndarray] = []
        episode_rewards: List[float] = []
        episode_dones: List[bool] = []
        episode_return: float = 0.0
        episode_length: int = 0

        if self.observation is None:
            # Reset env for the first function call
            self.observation = self.env.reset(seed=self.seed)
        elif not self.is_continuous:
            self.observation = self.env.reset()

        for current_step in range(num_samples):
            episode_observations.append(self.observation)

            action: np.ndarray = policy.get_action_numpy(self.observation)
            episode_actions.append(action)

            reward: float
            episode_done: bool
            self.observation, reward, episode_done, _ = self.env.step(action)

            episode_return += reward
            episode_rewards.append(reward)
            episode_dones.append(episode_done)

            episode_length += 1
            epoch_ended: bool = current_step == num_samples - 1

            if episode_done or epoch_ended:
                if epoch_ended and not episode_done:
                    logger.debug(
                        "The trajectory cut off at {} steps on the current episode".format(
                            episode_length
                        )
                    )

                episode_last_observation: np.ndarray = self.observation

                experience.observations.append(episode_observations)
                experience.actions.append(episode_actions)
                experience.rewards.append(episode_rewards)
                experience.last_observations.append(episode_last_observation)
                experience.dones.append(episode_dones)

                experience.episode_returns.append(episode_return)
                experience.episode_lengths.append(episode_length)

                if episode_done:
                    self.observation = self.env.reset()

                episode_return, episode_length = 0.0, 0
                (
                    episode_observations,
                    episode_actions,
                    episode_rewards,
                    episode_dones,
                ) = ([], [], [], [])

        return experience



from abc import ABC, abstractmethod

import numpy as np
from torch import Tensor, nn


class Policy(nn.Module, ABC):
    """
    Base policy class
    """

    @abstractmethod
    def get_action_tensor(self, observation: Tensor) -> Tensor:
        """
        Get action(s) given observation(s) from the environment

        :param observation: (Tensor) Observation(s) from the environment.
        :return: (Tensor) Action(s)
        """
        raise NotImplementedError

    @abstractmethod
    def get_action_numpy(self, observation: np.ndarray) -> np.ndarray:
        """
        Get action(s) given observation(s) from the environment

        :param observation: (np.ndarray) Observation(s) from the environment.
        :return: (np.ndarray) Action(s)
        """
        raise NotImplementedError



import numpy as np
import torch
from gym import Space
from torch import Tensor

from rl_replicas.policies import Policy


class RandomPolicy(Policy):
    """
    Random policy

    :param action_space: (Space) Action space.
    """

    def __init__(self, action_space: Space):
        self.action_space = action_space

    def get_action_tensor(self, observation: Tensor) -> Tensor:
        _ = observation  # Don't use observation
        action = self.action_space.sample()
        return torch.from_numpy(action)

    def get_action_numpy(self, observation: np.ndarray) -> np.ndarray:
        _ = observation  # Don't use observation
        action = self.action_space.sample()
        return action



"""
Policies
"""
from rl_replicas.policies.categorical_policy import CategoricalPolicy
from rl_replicas.policies.deterministic_policy import DeterministicPolicy
from rl_replicas.policies.gaussian_policy import GaussianPolicy
from rl_replicas.policies.policy import Policy
from rl_replicas.policies.random_policy import RandomPolicy
from rl_replicas.policies.stochastic_policy import StochasticPolicy



from typing import List, Optional, Tuple

import gym
import numpy as np

from rl_replicas.policies import Policy


class Evaluator:
    """
    Evaluator

    :param seed: (int) Seed.
    """

    def __init__(self, seed: Optional[int] = None):
        self.seed = seed

    def evaluate(
        self, policy: Policy, env: gym.Env, num_episodes: int
    ) -> Tuple[List[float], List[int]]:
        """
        Evaluate the policy running evaluation episodes.

        :param policy: (Policy) Policy.
        :param env: (gym.Env) Environment.
        :param num_episodes: (int) The number of episodes.
        """
        episode_returns: List[float] = []
        episode_lengths: List[int] = []

        observation: np.ndarray = env.reset(seed=self.seed)

        for _ in range(num_episodes):
            done: bool = False
            episode_return: float = 0.0
            episode_length: int = 0

            while not done:
                action: np.ndarray = policy.get_action_numpy(observation)

                reward: float
                observation, reward, done, _ = env.step(action)

                episode_return += reward
                episode_length += 1

            observation = env.reset()

            episode_returns.append(episode_return)
            episode_lengths.append(episode_length)

        return episode_returns, episode_lengths



import numpy as np
import torch
from torch import Tensor, nn
from torch.optim import Optimizer

from rl_replicas.policies.policy import Policy


class DeterministicPolicy(Policy):
    """
    Deterministic policy

    :param network: (nn.Module) Network.
    :param optimizer: (Optimizer) Optimizer.
    """

    def __init__(self, network: nn.Module, optimizer: Optimizer):
        super().__init__()

        self.network = network
        self.optimizer = optimizer

    def forward(self, observation: Tensor) -> Tensor:
        """
        Forward pass in policy

        :param observation: (Tensor) Observation of the environment.
        :return: (Tensor) Action(s).
        """
        action: Tensor = self.network(observation)

        return action

    def get_action_tensor(self, observation: Tensor) -> Tensor:
        with torch.no_grad():
            action: Tensor = self.forward(observation)

        return action

    def get_action_numpy(self, observation: np.ndarray) -> np.ndarray:
        observation_tensor: Tensor = torch.from_numpy(observation).float()
        with torch.no_grad():
            action: Tensor = self.forward(observation_tensor)

        return action.detach().numpy()
